@@ -1280,6 +1280,7 @@ neon_session_pool_destroy (NeonSessionPool *pool)
 	g_list_free (pool->unused_sessions);
 	
 	gnome_vfs_uri_unref (pool->uri);
+	g_free (pool);
 }
 
 
@@ -1558,6 +1559,8 @@ http_context_free (HttpContext *context)
 		neon_session_pool_insert (context->uri, context->session);
 		context->session = NULL;
 	}
+
+	g_free (context->path);
 	
 	gnome_vfs_uri_unref (context->uri);
 	
@@ -1573,20 +1576,20 @@ http_context_set_uri (HttpContext *context, GnomeVFSURI *uri)
 	if (context->uri)
 		gnome_vfs_uri_unref (context->uri);
 	
-	if (context->path == NULL)
+	if (context->path != NULL)
 		g_free (context->path);
 	
-	context->uri = gnome_vfs_uri_ref (uri);
+	context->uri = gnome_vfs_uri_dup (uri);
 	context->scheme = resolve_alias (gnome_vfs_uri_get_scheme (uri));
 	
-	if (gnome_vfs_uri_get_host_port (uri) == 0) {
+	if (gnome_vfs_uri_get_host_port (context->uri) == 0) {
 		if (g_str_equal (context->scheme, "https"))
-			gnome_vfs_uri_set_host_port (uri, DEFAULT_HTTPS_PORT);
+			gnome_vfs_uri_set_host_port (context->uri, DEFAULT_HTTPS_PORT);
 		else 
-			gnome_vfs_uri_set_host_port (uri, DEFAULT_HTTP_PORT);
+			gnome_vfs_uri_set_host_port (context->uri, DEFAULT_HTTP_PORT);
 	}
 	
-	uri_string = gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_USER_NAME
+	uri_string = gnome_vfs_uri_to_string (context->uri, GNOME_VFS_URI_HIDE_USER_NAME
 					      | GNOME_VFS_URI_HIDE_PASSWORD
 					      | GNOME_VFS_URI_HIDE_HOST_NAME
 					      | GNOME_VFS_URI_HIDE_HOST_PORT
@@ -1717,10 +1720,12 @@ http_get_file_info (HttpContext *context, GnomeVFSFileInfo *info, char **etag)
 	DEBUG_HTTP ("%d, %d", res, ne_get_status (req)->code);
 	ne_propfind_destroy (pfh);
 	
-	/* some server (eg. gws) colse the connection on an unknown command 
-	   so fall back to head for these too */
-	if (result != GNOME_VFS_ERROR_GENERIC  && 
-	    result != GNOME_VFS_ERROR_NOT_SUPPORTED) {
+	/* Let's be very cautious here! If the server doesn't respond with a 
+	   207 here just fall back to HEAD because some server server (eg. gws) 
+	   colse the connection on an unknown command, most (stupid) php scripts
+	   will treat PROPFIND as GET and some servers may deny us PROFIND but
+	   allow us HEAD. */
+	if (res == NE_OK && ne_get_status (req)->code == 207) {
 		
 		if (result == GNOME_VFS_OK) {
 			gnome_vfs_file_info_copy (info, pfctx.target);
@@ -1824,9 +1829,11 @@ http_list_directory (HttpContext *context, PropfindContext *pfctx)
 	pfh = ne_propfind_create (context->session, context->path, 1);
 	res = ne_propfind_named (pfh, file_info_props, propfind_result, pfctx);
 	
-	if (tofree != NULL)
+	if (tofree != NULL) {
 		g_free (tofree);
-	
+		tofree = NULL;
+	}
+
 	if (res == NE_REDIRECT) {
 		
 		ne_propfind_destroy (pfh);
@@ -1900,7 +1907,7 @@ end_response (void *userdata, void *response,
 }
 
 static int
-dav_request (ne_request *req)
+dav_request (ne_request *req, gboolean allow_redirect)
 {
 	ne_207_parser   *p207;
 	ne_xml_parser   *p;
@@ -1917,7 +1924,7 @@ dav_request (ne_request *req)
 	ne_207_set_propstat_handlers (p207, NULL, NULL /* end_propstat */);
     
 	ne_add_response_body_reader (req, ne_accept_207, ne_xml_parse_v, p);
-    
+	
 	res = ne_request_dispatch (req);
 	status = (ne_status *) ne_get_status (req);
     
@@ -1928,7 +1935,8 @@ dav_request (ne_request *req)
 			status->code = error;
 			status->klass = error % 100;
 		}
-	} else if (status->klass != 2) {
+	} else if (status->klass != 2 &&
+		   (!allow_redirect || res != NE_REDIRECT)) {
 		return NE_ERROR;
 	}
 	
@@ -2143,9 +2151,17 @@ get_start:
 	
 	if (IS_REDIRECT (status->code) || IS_AUTH_REQ (status->code)) {
 		/* send the body to /dev/null */				
-		while (ne_read_response_block (req, dropbuf, sizeof (dropbuf)))
+		while ((res = ne_read_response_block (req, dropbuf, sizeof (dropbuf))) > 0)
 			/* noop */;
 		
+		if (res < 0) {
+			handle->transfer_state = TRANSFER_ERROR;
+			result = GNOME_VFS_ERROR_IO;
+			handle->last_error = result;
+			ne_request_destroy (req);
+			return result;
+		}
+		
 		res = ne_end_request (req);
 		ne_request_destroy (req);
 		req = NULL;
@@ -2201,6 +2217,8 @@ do_open (GnomeVFSMethod 	*method,
 	HttpFileHandle *handle;
 	GnomeVFSResult result;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	_GNOME_VFS_METHOD_PARAM_CHECK (uri != NULL);
 	
@@ -2261,6 +2279,7 @@ do_open (GnomeVFSMethod 	*method,
 	
 	*method_handle = (GnomeVFSMethodHandle *) handle;
 	
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2279,6 +2298,8 @@ do_create (GnomeVFSMethod	 *method,
 	int res;
 	ne_request *req;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 
 	if ((mode & GNOME_VFS_OPEN_READ))
@@ -2331,6 +2352,7 @@ do_create (GnomeVFSMethod	 *method,
 
 	*method_handle = (GnomeVFSMethodHandle *) handle;	
 
+	DEBUG_HTTP_FUNC (0);
 	return result;
 	
 }
@@ -2344,6 +2366,8 @@ do_close (GnomeVFSMethod 	*method,
 	HttpFileHandle *handle;
 	HttpContext    *ctx;
 	int res;
+
+	DEBUG_HTTP_FUNC (1);
 	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	
@@ -2375,6 +2399,7 @@ do_close (GnomeVFSMethod 	*method,
 	
 	
 	http_file_handle_destroy (handle);
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2388,8 +2413,7 @@ do_read (GnomeVFSMethod 	*method,
 {
 	GnomeVFSResult result;
 	HttpFileHandle *handle;
-	GnomeVFSFileSize n;
-	int res;
+	ssize_t n;
 	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	
@@ -2412,30 +2436,37 @@ do_read (GnomeVFSMethod 	*method,
 	
 	n = ne_read_response_block (handle->transfer.read, buffer, num_bytes);
 
-	if (bytes_read != NULL)
-		*bytes_read = (n > 0 ? n : 0);
-	
 	if (n < 1) {
-		res  = ne_end_request (handle->transfer.read);
-		
+				
 		if (n == 0) {
+			ne_end_request (handle->transfer.read);
+			
 			result = GNOME_VFS_ERROR_EOF;
 			handle->transfer_state = TRANSFER_IDLE;
 		} else {
+			result = GNOME_VFS_ERROR_IO;
 			handle->transfer_state = TRANSFER_ERROR;
-			result = resolve_result (res, handle->transfer.read);
 		}
 		
 		ne_request_destroy (handle->transfer.read);
 		handle->transfer.read = NULL;
 		handle->last_error = result;
 		handle->offset = 0;
+		*bytes_read = 0;
+		
+		DEBUG_HTTP ("[read] error during read: %s", 
+				gnome_vfs_result_to_string (result));
+
 		return result;
 	}
 	
+		
+	/* cast is valid because n must be greater than 0 */
+	*bytes_read = n;
 
-	
-	handle->offset += num_bytes;
+	DEBUG_HTTP ("[read] bytes read %lld", *bytes_read);
+
+	handle->offset += *bytes_read;
 	return result;
 }
 
@@ -2510,6 +2541,8 @@ do_seek (GnomeVFSMethod	      *method,
 	HttpFileHandle 		*handle;
 	GnomeVFSFileOffset  	 new_position;
 
+	DEBUG_HTTP_FUNC (1);
+	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	
 	handle = (HttpFileHandle *) method_handle;
@@ -2558,13 +2591,10 @@ do_seek (GnomeVFSMethod	      *method,
 	handle->offset = new_position;
 	
 	if (handle->transfer_state == TRANSFER_READ) {
-				
 		http_transfer_abort (handle);
-		result = http_transfer_start (handle);
 	}
 	
-	
-	
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2613,6 +2643,8 @@ do_open_directory (GnomeVFSMethod 	    *method,
 	GnomeVFSResult  result;
 	HttpContext    *hctx;
 	PropfindContext *pfctx;
+
+	DEBUG_HTTP_FUNC (1);
 	
 	result = http_context_open  (uri, &hctx);
 	
@@ -2650,6 +2682,7 @@ do_open_directory (GnomeVFSMethod 	    *method,
 		*method_handle 	= (GnomeVFSMethodHandle  *) pfctx;
 	}
 	
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2660,6 +2693,7 @@ do_close_directory (GnomeVFSMethod 	 *method,
 {
 	PropfindContext *pfctx;
 	
+	DEBUG_HTTP_FUNC (1);
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 
 	pfctx = (PropfindContext *) method_handle;
@@ -2667,7 +2701,8 @@ do_close_directory (GnomeVFSMethod 	 *method,
 	propfind_context_clear (pfctx);
 	g_free (pfctx);
 	
-	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	DEBUG_HTTP_FUNC (0);
+	return GNOME_VFS_OK;
 }
 
 static GnomeVFSResult
@@ -2731,6 +2766,8 @@ do_get_file_info_from_handle (GnomeVFSMethod 		*method,
 	HttpFileHandle *handle;
 	GnomeVFSResult  result;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	
 	handle = (HttpFileHandle *) method_handle;
@@ -2772,6 +2809,8 @@ do_make_directory (GnomeVFSMethod  *method,
 	ne_request       *req;
 	int 		  res;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	uri_parent = gnome_vfs_uri_get_parent (uri);
 	
 	result = http_context_open (uri_parent, &hctx);
@@ -2790,11 +2829,17 @@ do_make_directory (GnomeVFSMethod  *method,
 	}
 	
 	http_context_set_uri (hctx, uri);
-	
+
+ mkcol_start:
 	req = ne_request_create (hctx->session, "MKCOL", hctx->path);
 	res = ne_request_dispatch (req);
 
-	if (res == NE_OK) {
+	if (res == NE_REDIRECT) {
+		result = http_follow_redirect (hctx);
+		
+		if (result == GNOME_VFS_OK)
+			goto mkcol_start;
+	} else if (res == NE_OK) {
 		const ne_status *status = ne_get_status (req);
 		if (status->code == 409)
 			result = GNOME_VFS_ERROR_NOT_FOUND;
@@ -2809,6 +2854,7 @@ do_make_directory (GnomeVFSMethod  *method,
  out:	
 	gnome_vfs_uri_unref (uri_parent);
 	http_context_free (hctx);
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2823,6 +2869,7 @@ do_remove_directory (GnomeVFSMethod  *method,
 	ne_request     *req;
 	int  		res;
 	
+	DEBUG_HTTP_FUNC (1);
 	result = http_context_open  (uri, &hctx);
 	
 	if (result != GNOME_VFS_OK)
@@ -2845,7 +2892,7 @@ do_remove_directory (GnomeVFSMethod  *method,
 	}
 	
 	req = ne_request_create (hctx->session, "DELETE", hctx->path);
-	res = dav_request (req);
+	res = dav_request (req, FALSE);
 	
 	result = resolve_result (res, req);
 	ne_request_destroy (req);
@@ -2853,9 +2900,21 @@ do_remove_directory (GnomeVFSMethod  *method,
  out:	
 	propfind_context_clear (&pfctx);
 	http_context_free (hctx);
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
+static GnomeVFSURI *
+resolve_schema_alias (GnomeVFSURI *uri)
+{
+	GnomeVFSURI *resolved;
+
+	resolved = gnome_vfs_uri_dup (uri);
+	g_free (resolved->method_string);
+	resolved->method_string = g_strdup (resolve_alias (gnome_vfs_uri_get_scheme (uri)));
+	return resolved;
+}
+
 
 static GnomeVFSResult
 do_move (GnomeVFSMethod  *method,
@@ -2869,8 +2928,10 @@ do_move (GnomeVFSMethod  *method,
 	char	    *dest;
 	ne_request  *req;
 	int 	     res;
-
+	GnomeVFSURI *dest_uri;
 	
+	DEBUG_HTTP_FUNC (1);
+
 	if (! http_session_uri_equal (old_uri, new_uri)) {
 		return GNOME_VFS_ERROR_NOT_SAME_FILE_SYSTEM;
 	}	
@@ -2879,22 +2940,34 @@ do_move (GnomeVFSMethod  *method,
 	
 	if (result != GNOME_VFS_OK)
 		return result;
-	
-	dest = gnome_vfs_uri_to_string (new_uri, GNOME_VFS_URI_HIDE_USER_NAME | 
+
+	dest_uri = resolve_schema_alias (new_uri);
+	dest = gnome_vfs_uri_to_string (dest_uri, GNOME_VFS_URI_HIDE_USER_NAME | 
 					GNOME_VFS_URI_HIDE_PASSWORD);
+	gnome_vfs_uri_unref (dest_uri);
 	
 
+ head_start:
 	req = ne_request_create (hctx->session, "MOVE", hctx->path);
 	
 	ne_add_request_header (req, "Destination", dest);
     	ne_add_request_header (req, "Overwrite", force_replace ? "T" : "F");
-
 	
-	res = dav_request (req);
-	result = resolve_result (res, req);
+	
+	res = dav_request (req, TRUE);
+	if (res == NE_REDIRECT) {
+		result = http_follow_redirect (hctx);
+
+		/* TODO: Do we leak req here? */
+		if (result == GNOME_VFS_OK)
+			goto head_start;
+	} else {
+		result = resolve_result (res, req);
+	}
 
 	http_context_free (hctx);
 	
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2909,6 +2982,7 @@ do_unlink (GnomeVFSMethod  *method,
 	ne_request       *req;
 	int 		  res;
 	
+	DEBUG_HTTP_FUNC (1);
 	result = http_context_open (uri, &hctx);
 	
 	if (result != GNOME_VFS_OK)
@@ -2927,7 +3001,7 @@ do_unlink (GnomeVFSMethod  *method,
 	}
 	
 	req = ne_request_create (hctx->session, "DELETE", hctx->path);
-	res = dav_request (req);
+	res = dav_request (req, FALSE);
 
 	result = resolve_result (res, req);
 	ne_request_destroy (req);
@@ -2935,6 +3009,7 @@ do_unlink (GnomeVFSMethod  *method,
  out:	
 	http_context_free (hctx);
 	gnome_vfs_file_info_unref (file_info);
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2961,6 +3036,8 @@ do_set_file_info (GnomeVFSMethod 		*method,
 	GnomeVFSURI *parent_uri, *new_uri;
 	GnomeVFSResult result;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	if ((mask & ~(GNOME_VFS_SET_FILE_INFO_NAME)) != 0) {
 		return GNOME_VFS_ERROR_NOT_SUPPORTED;
 	}
@@ -2982,6 +3059,8 @@ do_set_file_info (GnomeVFSMethod 		*method,
 	
 	gnome_vfs_uri_unref (new_uri);
 	
+	DEBUG_HTTP_FUNC (0);
+	
 	return result;
 }
 
@@ -3079,6 +3158,7 @@ GnomeVFSMethod *
 vfs_module_init (const char *method_name, const char *args)
 {
 	proxy_init ();
+	/* ne_debug_init (stdout, 0xfffe); */
 	neon_session_pool_init ();
 	http_auth_cache_init ();
 	quick_allow_lookup_init ();
